import json
import ssl
import time
from contextlib import contextmanager

import boto3
import psycopg2
import pymysql
import redis
from pymemcache import PooledClient
from pymysql import OperationalError

from common.secretmanager import SecretManager

context = ssl.create_default_context()


class RDSConnection:
    def __init__(self, db_secret_name, access_key=None, secret_key=None, region_name=None):
        self.__region_name = region_name
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__db_secret_name = db_secret_name
        self.__aws_secret_manager = SecretManager(region_name=self.__region_name, access_key=self.__access_key,
                                                  secret_key=self.__secret_key)
        self.__database_values = self.__aws_secret_manager.get_secret_key_value(self.__db_secret_name)
        self.__user = self.__database_values.get('username')
        self.__password = self.__database_values.get('password')
        self.__host = self.__database_values.get('host')
        self.__port = int(self.__database_values.get('port')) if self.__database_values.get('port') else 3306
        self.__conn = None
        self.__connect_timeout = 30

    @contextmanager
    def get_connection(self):
        """ Context manager to handle connection management """
        if self.__conn is None or self.__conn.open is False:
            self.__conn = pymysql.connect(user=self.__user, password=self.__password, host=self.__host,
                                          port=self.__port, connect_timeout=self.__connect_timeout)
        try:
            yield self.__conn
        finally:
            # Only close if it was opened here
            if self.__conn and self.__conn.open:
                self.__conn.close()

    def conn(self):
        return pymysql.connect(user=self.__user, password=self.__password, host=self.__host, port=self.__port,
                               connect_timeout=self.__connect_timeout)

    @staticmethod
    def cursor(conn):
        """ Return the cursor object """
        return conn.cursor()

    def executemany_query(self, query, values=None, retry_count=3):
        """ Execute the query with multiple values """
        attempt = 0
        while attempt < retry_count:
            try:
                with self.get_connection() as conn:
                    cursor = self.cursor(conn=conn)
                    cursor.executemany(query, values)
                    conn.commit()
                    return cursor
            except OperationalError as error:
                if error.args[0] in (1213, 1205):  # This is for deadlock error
                    attempt += 1
                    wait_time = 2 ** attempt
                    print(f"INFO: Deadlock detected. retry in {wait_time}")
                    time.sleep(wait_time)
                else:
                    raise
            except Exception as e:
                print(f"ERROR: {str(e)}")
                raise
        raise Exception(f"ERROR: Failed to execute query after {retry_count} attempts")

    def execute_query(self, query, values=None, retry_count=3):
        """ Execute a single query """
        attempt = 0
        while attempt < retry_count:
            try:
                with self.get_connection() as conn:
                    cursor = self.cursor(conn=conn)
                    cursor.execute(query, values)
                    conn.commit()
                    return cursor
            except OperationalError as error:
                if error.args[0] in (1213, 1205):  # This is for deadlock error
                    attempt += 1
                    wait_time = 2 ** attempt
                    print(f"INFO: Deadlock detected. retry in {wait_time}")
                    time.sleep(wait_time)
                else:
                    raise
            except Exception as e:
                print(f"ERROR: {str(e)}")
                raise
        raise Exception(f"ERROR: Failed to execute query after {retry_count} attempts")

    def fetchall(self, query, values=None):
        """ Fetch all rows for a given query """
        try:
            with self.get_connection() as conn:
                cursor = self.cursor(conn=conn)
                cursor.execute(query, values)
                return cursor.fetchall()
        except Exception as e:
            print(f'ERROR: {str(e)}')
            return []


class RedisConnection:
    def __init__(self, redis_secret_name, access_key=None, secret_key=None, region_name=None):
        self.__region_name = region_name
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__redis_secret_name = redis_secret_name
        self.__aws_secret_manager = SecretManager(region_name=self.__region_name, access_key=self.__access_key,
                                                  secret_key=self.__secret_key)
        self.__redis_values = self.__aws_secret_manager.get_secret_key_value(self.__redis_secret_name)
        self.__host = self.__redis_values.get('host')
        self.__port = self.__redis_values.get('port')

    def conn(self):
        try:
            red_conn = redis.StrictRedis(host=self.__host, port=self.__port)
            return red_conn
        except redis.ConnectionError as e:
            print(f"Redis Connection Error: {e}")
        except redis.RedisError as e:
            print(f"Redis Error: {e}")

    def get(self, key):
        conn = self.conn()
        return conn.get(key)

    def set(self, key, data, ex_time):
        conn = self.conn()
        conn.set(key, data, ex=ex_time)


class TimescaleDBConnection:
    def __init__(self, timescaledb_secret_name, access_key=None, secret_key=None, region_name=None):
        self.__region_name = region_name
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__timescaledb_secret_name = timescaledb_secret_name
        self.__aws_secret_manager = SecretManager(region_name=self.__region_name, access_key=self.__access_key,
                                                  secret_key=self.__secret_key)
        self.__timescaledb_values = self.__aws_secret_manager.get_secret_key_value(self.__timescaledb_secret_name)
        self.__username = self.__timescaledb_values.get('username')
        self.__password = self.__timescaledb_values.get('password')
        self.__host = self.__timescaledb_values.get('host')
        self.__port = int(self.__timescaledb_values.get('port'))
        self.__dbname = self.__timescaledb_values.get('dbname')

    def conn(self):
        return psycopg2.connect(user=self.__username, password=self.__password, host=self.__host, port=self.__port,
                                database=self.__dbname)

    def cursor(self, conn):
        """ Return a cursor object """
        return conn.cursor()

    def execute_query(self, query, values=None):
        """ Execute a single query """
        conn = self.conn()
        try:
            cursor = self.cursor(conn=conn)
            cursor.execute(query, values)
            conn.commit()
            return cursor
        finally:
            conn.close()

    def executemany_query(self, query, values=None):
        """ Execute a query with multiple parameters """
        conn = self.conn()
        try:
            cursor = self.cursor(conn=conn)
            cursor.executemany(query, values)
            conn.commit()
            return cursor
        finally:
            conn.close()

    def fetchall(self, query, values=None):
        """ Execute a query and fetch all data """
        conn = self.conn()
        try:
            cursor = self.cursor(conn=conn)
            cursor.execute(query, values)
            result = cursor.fetchall()
            return result
        except Exception as e:
            print(f'ERROR: {str(e)}')
            return []
        finally:
            conn.close()


class MemcachedConnection:
    def __init__(self, memcached_secret_name, access_key=None, secret_key=None, region_name=None, add_context=False):
        self.__region_name = region_name
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__memcached_secret_name = memcached_secret_name
        self.__aws_secret_manager = SecretManager(region_name=self.__region_name, access_key=self.__access_key,
                                                  secret_key=self.__secret_key)
        self.__memcached_values = self.__aws_secret_manager.get_secret_key_value(self.__memcached_secret_name)
        self.__host = self.__memcached_values.get('host')
        self.__port = self.__memcached_values.get('port')
        if add_context:
            self.__client = PooledClient((self.__host, self.__port), tls_context=context, max_pool_size=20)
        else:
            self.__client = PooledClient((self.__host, self.__port), max_pool_size=20)

    def set_data_into_memcache(self, key, data, ttl=600, verbose=False):
        try:
            key = key.replace(" ", "")
            json_data = json.dumps(data, default=str)

            self.__client.set(key, json_data, expire=ttl)
            if verbose:
                print(f"INFO: Data set to memcached: {data}")
        except Exception as error:
            print(f"ERROR: {error}")

    def get_data_from_memcached(self, key, verbose=False):
        try:
            key = key.replace(" ", "")
            cached_data = self.__client.get(key)

            if cached_data:
                data = json.loads(cached_data)
                if verbose:
                    print(f"INFO: Data got from memcached: {data}")
                return data, True
        except Exception as error:
            print(f"ERROR: {error}")

        return None, None

    def delete_data_from_memcache(self, key, verbose=False):
        try:
            key = key.replace(" ", "")
            self.__client.delete(key)
            if verbose:
                print(f"INFO: Data deleted from memcached for key: {key}")
        except Exception as error:
            print(f"ERROR: {error}")


class TimestreamConnection:
    def __init__(self, access_key=None, secret_key=None, region_name=None):
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__region_name = region_name

    def get_write_client(self):
        try:
            session = boto3.session.Session(aws_access_key_id=self.__access_key,
                                            aws_secret_access_key=self.__secret_key)
            return session.client(service_name="timestream-write", region_name=self.__region_name)
        except Exception as e:
            print(f'Exception while getting write client for timestream: {e}')
            raise e

    def get_query_client(self):
        try:
            session = boto3.session.Session(aws_access_key_id=self.__access_key,
                                            aws_secret_access_key=self.__secret_key)
            return session.client(service_name="timestream-query", region_name=self.__region_name)
        except Exception as e:
            print(f'Exception while getting query client for timestream: {e}')
            raise e



from datetime import datetime

from blockchain import read_db
from blockchain.constants import GET_ACTIVE_SYMBOLS_DATA_QUERY, GET_AUTHOR_HANDLE_DETAILS_FROM_TWITTER_PROFILE_QUERY, \
    SEARCH_TELEGRAM_DATA_QUERY, GET_TWEETS_QUERY, GET_AUTHOR_HANDLE_DETAILS_FROM_TWITTER_PROFILE_EXACT_MATCH_QUERY, \
    SEARCH_TELEGRAM_DATA_EXACT_MATCH_QUERY, GET_ACTIVE_COINS_DATA_QUERY, GET_CONFIGURATION_FOR_SEARCH_QUERY
from common_utils.chain_utils import chain_to_id
from common_utils.time_utils import format_datetime


def get_onchain_data(search_term, limit=3, start=0, is_exact_match=False):
    result = []
    total_results = 0

    if len(search_term) >= 40:
        search_condition = f"token_id = '{search_term}' OR pair_id = '{search_term}'"
    elif len(search_term) < 40 and is_exact_match:
        search_condition = f"symbol = '{search_term}' OR name = '{search_term}'"
    else:
        search_condition = f"symbol LIKE '{search_term}%' OR symbol LIKE '${search_term}%' OR name LIKE '{search_term}%' OR name LIKE ' {search_term}%'"

    if start == 0:
        coin_data = read_db.fetchall(query=GET_ACTIVE_COINS_DATA_QUERY.format(search_condition=search_condition))

        for cd in coin_data:
            (symbol, name, is_coin, chain_id, token_id, pair_id, vol_24_hr, liquidity, marketcap, icon, buy_tax,
             sell_tax, pair_created_at, twitter, telegram, website) = cd
            if not pair_created_at:
                age_in_seconds = None
            else:
                age_in_seconds = datetime.utcnow().timestamp() - int(pair_created_at)
            result.append({
                "symbol": symbol,  # will remove this
                "token_symbol": symbol,
                "name": name,  # will remove this
                "token_name": name,
                "is_coin": is_coin,
                "chain_id": None,
                "chain": None,
                "token_id": token_id,
                "pair_id": pair_id,
                "vol_24_hr": vol_24_hr,  # will remove this
                "vol_24hr": vol_24_hr,
                "liquidity": liquidity,
                "marketcap": marketcap,  # will remove this
                "market_cap": marketcap,
                "icon": icon,  # will remove this
                "token_icon": icon,
                "buy_tax": buy_tax,
                "sell_tax": sell_tax,
                "age_in_seconds": age_in_seconds,
                "pair_created_at": pair_created_at,
                "twitter": twitter,  # will remove this
                "token_twitter": twitter,
                "telegram": telegram,  # will remove this
                "token_telegram": telegram,
                "website": website,  # will remove this
                "token_website": website,
                "is_honeypot": None,
                "can_mint": None,
                "is_proxy": None,
                "is_blacklisted": None,
                "can_burn": None,
                "is_scam": None,
                "can_freeze": None,
                "is_contract_verified": None,
                "pc_24_hr": None
            })
            total_results += 1

        if len(result) >= limit:
            return True, (result[:limit], limit)
    
    internal_search_limit = read_db.fetchall(query=GET_CONFIGURATION_FOR_SEARCH_QUERY)[0][0]

    query = GET_ACTIVE_SYMBOLS_DATA_QUERY.format(search_condition=search_condition, search_term=search_term,
                                                 internal_search_limit=internal_search_limit, start=start,
                                                 limit=limit-len(result))

    if data := read_db.fetchall(query=query):
        for i in data:
            (symbol, name, is_coin, chain, token_id, pair_id, vol_24_hr, liquidity, marketcap, icon, buy_tax, sell_tax,
             pair_created_at, twitter, telegram, website, order_number, score, is_honeypot, can_mint, is_proxy,
             is_blacklisted, can_burn, is_scam, can_freeze, is_contract_verified, pc_24_hr, match_priority) = i
            chain = None if is_coin == 2 else chain
            is_coin = is_coin if not is_coin else 1
            chain_id = chain_to_id.get(chain)
            if not pair_created_at:
                age_in_seconds = None
            else:
                age_in_seconds = datetime.utcnow().timestamp() - int(pair_created_at)
            result.append({
                "symbol": symbol,   # will remove this
                "token_symbol": symbol,
                "name": name,   # will remove this
                "token_name": name,
                "is_coin": is_coin,
                "chain_id": chain_id,
                "chain": chain,
                "token_id": token_id,
                "pair_id": pair_id,
                "vol_24_hr": vol_24_hr, # will remove this
                "vol_24hr": vol_24_hr,
                "liquidity": liquidity,
                "marketcap": marketcap, # will remove this
                "market_cap": marketcap,
                "icon": icon,   # will remove this
                "token_icon": icon,
                "buy_tax": buy_tax,
                "sell_tax": sell_tax,
                "age_in_seconds": age_in_seconds,
                "pair_created_at": pair_created_at,
                "twitter": twitter, # will remove this
                "token_twitter": twitter,
                "telegram": telegram,   # will remove this
                "token_telegram": telegram,
                "website": website, # will remove this
                "token_website": website,
                "is_honeypot": is_honeypot,
                "can_mint": can_mint,
                "is_proxy": is_proxy,
                "is_blacklisted": is_blacklisted,
                "can_burn": can_burn,
                "is_scam": is_scam,
                "can_freeze": can_freeze,
                "is_contract_verified": is_contract_verified,
                "pc_24_hr": pc_24_hr
            })
            total_results += 1
        return True, (result, total_results)
    else:
        return False, "No match found!"


def get_twitter_author_handle_data(search_term, limit=3, start=0, is_exact_match=False):
    result = []
    total_results = 0

    if is_exact_match:
        query = GET_AUTHOR_HANDLE_DETAILS_FROM_TWITTER_PROFILE_EXACT_MATCH_QUERY.format(author_handle=search_term,
                                                                                        start=start, limit=limit)
    else:
        query = GET_AUTHOR_HANDLE_DETAILS_FROM_TWITTER_PROFILE_QUERY.format(author_handle=search_term, start=start,
                                                                            limit=limit)

    if details_from_twitter_profile := read_db.fetchall(query=query):
        for dftp in details_from_twitter_profile:
            (author_handle, name, bio, url_in_bio, profile_image_url, profile_banner_url, followers_count,
             followings_count, account_created_at, lifetime_tweets, lifetime_views, engagement_score, followers_impact,
             symbol_mentions, total_mentions, symbols_in_last_24hr, new_symbols_in_last_24hr,
             unique_author_handle_mentions, author_handle_mentions, tags, mindshare) = dftp

            result.append({
                "author_handle": author_handle,
                "name": name,
                "bio": bio,
                "url_in_bio": url_in_bio,
                "profile_image_url": profile_image_url,
                "profile_banner_url": profile_banner_url,
                "followers_count": followers_count,
                "followings_count": followings_count,
                "account_created_at": account_created_at,
                "crypto_tweets_all": lifetime_tweets,
                "crypto_tweets_views_all": lifetime_views,
                "engagement_score": round(engagement_score) if engagement_score else engagement_score,
                "followers_impact": round(followers_impact) if followers_impact else followers_impact,
                "total_symbols_mentioned": symbol_mentions, # this the total unique symbol mentioned by the author handle
                "symbols_mentioned_in_last_24hr": symbols_in_last_24hr,
                "new_symbols_mentioned_in_last_24hr": new_symbols_in_last_24hr,
                "unique_author_handle_count": unique_author_handle_mentions,
                "total_mention_count": total_mentions,  # this is total mentions done by author handle
                "tags": tags.split(",") if tags else [],
                "mentions_24hr": author_handle_mentions,    # this is the total mention of author handle done by other users
                "mindshare": mindshare
            })
            total_results += 1
        return True, (result, total_results)
    else:
        return False, "No match found!"


def get_telegram_data(search_term, limit=3, start=0, is_exact_match=False):
    result = []
    total_results = 0

    if is_exact_match:
        query = SEARCH_TELEGRAM_DATA_EXACT_MATCH_QUERY.format(search_term=search_term, limit=limit, start=start)
    else:
        query = SEARCH_TELEGRAM_DATA_QUERY.format(search_term=search_term, limit=limit, start=start)

    if telegram_filter_data := read_db.fetchall(query=query):
        for tfd in telegram_filter_data:
            (channel_id, total_mentions, token_mentions, average_mentions_per_day, name, image_url, tg_link,
             members_count, channel_age, win_rate_30_day) = tfd
            telegram_response_dict = {
                "channel_id": channel_id,
                "channel_name": name,
                "image_url": image_url,
                "channel_link": tg_link,
                "total_mentions": total_mentions,
                "token_mentions": token_mentions,
                "members_count": members_count,
                "channel_age": str(channel_age.timestamp()) if channel_age else None,
                "average_mentions_per_day": average_mentions_per_day,
                "win_rate": win_rate_30_day
            }
            result.append(telegram_response_dict)
            total_results += 1
        return True, (result, total_results)
    else:
        return False, "No match found!"


def get_tweets_data(search_term=None, tweet_id=None, author_handle=None, limit=3, start=0):
    result = []
    if not search_term and not tweet_id and not author_handle:
        return False, "Either search_term, tweet_id or author_handle is required"

    where_condition = ""
    if search_term:
        where_condition = f"body LIKE '%{search_term}%'"
    elif tweet_id:
        where_condition = f"tweet_id = '{tweet_id}'"
    elif author_handle:
        where_condition = f"author_handle = '{author_handle}'"
    
    query = GET_TWEETS_QUERY.format(where_condition=where_condition, start=start, limit=limit)

    if data := read_db.fetchall(query=query):
        for i in data:
            tweet_id, body, author_handle, tweet_create_time = i
            result.append({
                "tweet_id": tweet_id,
                "tweet_body": body,
                "author_handle": author_handle,
                "tweet_create_time": format_datetime(input_date_string=str(tweet_create_time))
            })

        return True, result
    else:
        return False, "No match found!"



def abbreviate_number(number):
    abbreviations = [
        (1e9, 'B'),
        (1e6, 'M'),
        (1e3, 'K')
    ]
    if not number:
        return 'N/A'
    else:
        number = float(number)

        if number < 1000:
            return str(int(number))

        for factor, suffix in abbreviations:
            if number >= factor:
                if number == factor:  # Exactly 1 billion (or other thresholds)
                    abbreviated = f"1{suffix}"
                else:
                    abbreviated = f"{number / factor:.1f}{suffix}"

                return abbreviated

        return str(number)


from setuptools import setup, find_packages

with open("./README.md") as readme:
    long_description = readme.read()

setup(
    name="cryptoowl",
    version="0.1.1",
    author="Cryptoowl",
    author_email="cryptoowl.app@gmail.com",
    description="A library, that stores commonly used code for different modules in the CryptoOwl application",
    long_description=long_description,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pytz",
        "boto3",
        "PyMySQL",
        "redis",
        "botocore",
        "setuptools",
        "pymemcache",
        "psycopg2-binary",
        "requests",
        "pydantic"
    ],
)
